from ariadne import gql

from .base_type import BaseType
from .convert_case import convert_case
from .deferred_type import DeferredType
from .directive_type import DirectiveType
from .enum_type import EnumType
from .executable_schema import make_executable_schema
from .input_type import InputType
from .interface_type import InterfaceType
from .mutation_type import MutationType
from .object_type import ObjectType
from .scalar_type import ScalarType
from .subscription_type import SubscriptionType
from .union_type import UnionType
from .utils import create_alias_resolver, parse_definition

__all__ = [
    "BaseType",
    "DeferredType",
    "DirectiveType",
    "EnumType",
    "InputType",
    "InterfaceType",
    "MutationType",
    "ObjectType",
    "ScalarType",
    "SubscriptionType",
    "UnionType",
    "convert_case",
    "create_alias_resolver",
    "gql",
    "make_executable_schema",
    "parse_definition",
]



from typing import Callable, Dict, Optional, Union

from graphql import GraphQLFieldResolver

from .types import FieldsDict
from .utils import create_alias_resolver

Aliases = Dict[str, str]
FieldsArgs = Dict[str, Dict[str, str]]


class ResolversMixin:
    """Adds aliases and resolvers logic to GraphQL type"""

    __aliases__: Optional[Union[Aliases, Callable[..., Aliases]]] = None
    __fields_args__: Optional[Union[FieldsArgs, Callable[..., FieldsArgs]]] = None

    graphql_name: str
    graphql_fields: FieldsDict

    resolvers: Dict[str, GraphQLFieldResolver]

    @classmethod
    def __validate_aliases__(cls):
        if not cls.__aliases__:
            return

        invalid_aliases = set(cls.__aliases__) - set(cls.graphql_fields)
        if invalid_aliases:
            raise ValueError(
                f"{cls.__name__} class was defined with aliases for fields not in "
                f"GraphQL type: {', '.join(invalid_aliases)}"
            )

    @classmethod
    def __validate_fields_args__(cls):
        if not cls.__fields_args__:
            return

        invalid_fields = set(cls.__fields_args__) - set(cls.graphql_fields)
        if invalid_fields:
            raise ValueError(
                f"{cls.__name__} class was defined with fields args mappings "
                f"for fields not in GraphQL type: {', '.join(invalid_fields)}"
            )

        for field_name, field_args in cls.__fields_args__.items():
            defined_args = [
                arg.name.value for arg in cls.graphql_fields[field_name].arguments
            ]
            invalid_args = set(field_args) - set(defined_args)
            if invalid_args:
                raise ValueError(
                    f"{cls.__name__} class was defined with args mappings not in "
                    f"not in '{field_name}' field: {', '.join(invalid_args)}"
                )

    @classmethod
    def __get_resolvers__(cls):
        aliases = cls.__aliases__ or {}
        defined_resolvers = cls.__get_defined_resolvers__()

        used_resolvers = []
        resolvers = {}

        for field_name in cls.graphql_fields:
            if aliases and field_name in aliases:
                resolver_name = aliases[field_name]
                if resolver_name in defined_resolvers:
                    used_resolvers.append(resolver_name)
                    resolvers[field_name] = defined_resolvers[resolver_name]
                else:
                    resolvers[field_name] = create_alias_resolver(resolver_name)

            elif field_name in defined_resolvers:
                used_resolvers.append(field_name)
                resolvers[field_name] = defined_resolvers[field_name]

        unused_resolvers = [
            f"resolve_{field_name}"
            for field_name in set(defined_resolvers) - set(used_resolvers)
        ]
        if unused_resolvers:
            raise ValueError(
                f"{cls.__name__} class was defined with resolvers for fields not in "
                f"GraphQL type: {', '.join(unused_resolvers)}"
            )

        return resolvers

    @classmethod
    def __get_defined_resolvers__(cls) -> Dict[str, Callable]:
        resolvers = {}
        for name in dir(cls):
            if not name.startswith("resolve_"):
                continue

            value = getattr(cls, name)
            if callable(value):
                resolvers[name[8:]] = value

        return resolvers



from typing import Dict, Optional, Union, cast

from ariadne import convert_camel_case_to_snake
from graphql import DefinitionNode

from .types import FieldsDict

Overrides = Dict[str, str]
ArgsOverrides = Dict[str, Overrides]


def convert_case(
    overrides_or_fields: Optional[Union[FieldsDict, dict]] = None,
    map_fields_args=False,
):
    no_args_call = convert_case_call_without_args(overrides_or_fields)

    overrides = {}
    if not no_args_call:
        overrides = cast(dict, overrides_or_fields)

    def create_case_mappings(fields: FieldsDict, map_fields_args=False):
        if map_fields_args:
            return convert_args_cas(fields, overrides)

        return convert_aliases_case(fields, overrides)

    if no_args_call:
        fields = cast(FieldsDict, overrides_or_fields)
        return create_case_mappings(fields, map_fields_args)

    return create_case_mappings


def convert_case_call_without_args(
    overrides_or_fields: Optional[Union[FieldsDict, dict]] = None
) -> bool:
    if overrides_or_fields is None:
        return True

    if isinstance(list(overrides_or_fields.values())[0], DefinitionNode):
        return True

    return False


def convert_aliases_case(fields: FieldsDict, overrides: Overrides) -> Overrides:
    final_mappings = {}
    for field_name in fields:
        if field_name in overrides:
            field_name_final = overrides[field_name]
        else:
            field_name_final = convert_camel_case_to_snake(field_name)
        if field_name != field_name_final:
            final_mappings[field_name] = field_name_final
    return final_mappings


def convert_args_cas(fields: FieldsDict, overrides: ArgsOverrides) -> ArgsOverrides:
    final_mappings = {}
    for field_name, field_def in fields.items():
        arg_overrides: Overrides = overrides.get(field_name, {})
        arg_mappings = {}
        for arg in field_def.arguments:
            arg_name = arg.name.value
            if arg_name in arg_overrides:
                arg_name_final = arg_overrides[arg_name]
            else:
                arg_name_final = convert_camel_case_to_snake(arg_name)
            if arg_name != arg_name_final:
                arg_mappings[arg_name] = arg_name_final
        if arg_mappings:
            final_mappings[field_name] = arg_mappings
    return final_mappings



from typing import List, Type

from graphql import DefinitionNode, GraphQLSchema

from .dependencies import Dependencies
from .types import RequirementsDict


class BaseType:
    __abstract__ = True
    __schema__: str
    __requires__: List[Type["BaseType"]] = []

    graphql_name: str
    graphql_type: Type[DefinitionNode]

    @classmethod
    def __get_requirements__(cls) -> RequirementsDict:
        return {req.graphql_name: req.graphql_type for req in cls.__requires__}

    @classmethod
    def __validate_requirements__(
        cls, requirements: RequirementsDict, dependencies: Dependencies
    ):
        for graphql_name in dependencies:
            if graphql_name not in requirements:
                raise ValueError(
                    f"{cls.__name__} class was defined without required GraphQL "
                    f"definition for '{graphql_name}' in __requires__"
                )

    @classmethod
    def __bind_to_schema__(cls, schema: GraphQLSchema):
        raise NotImplementedError()



from typing import Type, cast

from ariadne import SchemaDirectiveVisitor
from graphql import (
    DefinitionNode,
    DirectiveDefinitionNode,
)

from .base_type import BaseType
from .utils import parse_definition


class DirectiveType(BaseType):
    __abstract__ = True
    __visitor__: Type[SchemaDirectiveVisitor]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)

        cls.__validate_visitor__()

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> DirectiveDefinitionNode:
        if not isinstance(type_def, DirectiveDefinitionNode):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ "
                "without GraphQL directive"
            )

        return cast(DirectiveDefinitionNode, type_def)

    @classmethod
    def __validate_visitor__(cls):
        if not getattr(cls, "__visitor__", None):
            raise AttributeError(
                f"{cls.__name__} class was defined without __visitor__ attribute"
            )

    @staticmethod
    def __bind_to_schema__(*_):
        pass  # Binding directive to schema is noop



from enum import Enum
from typing import List, Optional, Type, Union, cast

import ariadne
from graphql import (
    DefinitionNode,
    GraphQLSchema,
    EnumTypeDefinitionNode,
    EnumTypeExtensionNode,
)

from .base_type import BaseType
from .types import RequirementsDict
from .utils import parse_definition

EnumNodeType = Union[EnumTypeDefinitionNode, EnumTypeExtensionNode]


class EnumType(BaseType):
    __abstract__ = True
    __enum__: Optional[Union[Type[Enum], dict]] = None

    graphql_type: Union[Type[EnumTypeDefinitionNode], Type[EnumTypeExtensionNode]]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

        values = cls.__get_values__(graphql_def)
        cls.__validate_values__(values)

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> EnumNodeType:
        if not isinstance(type_def, (EnumTypeDefinitionNode, EnumTypeExtensionNode)):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ without GraphQL enum"
            )

        return cast(EnumNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: EnumNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, EnumTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} graphql type was defined without required GraphQL "
                f"type definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != EnumTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL enum "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __get_values__(cls, type_def: EnumNodeType) -> List[str]:
        if not type_def.values and not (
            isinstance(type_def, EnumTypeExtensionNode) and type_def.directives
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing "
                f"empty, GraphQL enum definition"
            )

        return [value.name.value for value in type_def.values]

    @classmethod
    def __validate_values__(cls, values: List[str]):
        if not cls.__enum__:
            return

        if isinstance(cls.__enum__, dict):
            enum_keys = set(cls.__enum__.keys())
        else:
            enum_keys = set(cls.__enum__.__members__.keys())

        missing_keys = set(values) - enum_keys
        if missing_keys:
            raise ValueError(
                f"{cls.__name__} class was defined with __enum__ missing following "
                f"items required by GraphQL definition: {', '.join(missing_keys)}"
            )

        extra_keys = enum_keys - set(values)
        if extra_keys:
            raise ValueError(
                f"{cls.__name__} class was defined with __enum__ containing extra "
                f"items missing in GraphQL definition: {', '.join(extra_keys)}"
            )

    @classmethod
    def __bind_to_schema__(cls, schema: GraphQLSchema):
        if cls.__enum__:
            bindable = ariadne.EnumType(cls.graphql_name, cls.__enum__)
            bindable.bind_to_schema(schema)

    @classmethod
    def __bind_to_default_values__(cls, schema: GraphQLSchema):
        if cls.__enum__:
            bindable = ariadne.EnumType(cls.graphql_name, cls.__enum__)
            bindable.bind_to_default_values(schema)



from typing import Callable, Dict, Union, cast

from graphql import (
    DefinitionNode,
    GraphQLFieldResolver,
    GraphQLObjectType,
    GraphQLSchema,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
)

from .object_type import ObjectType

ObjectNodeType = Union[ObjectTypeDefinitionNode, ObjectTypeExtensionNode]


class SubscriptionType(ObjectType):
    __abstract__ = True

    subscribers: Dict[str, GraphQLFieldResolver]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.subscribers = cls.__get_subscribers__()

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> ObjectNodeType:
        if not isinstance(
            type_def, (ObjectTypeDefinitionNode, ObjectTypeExtensionNode)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ without GraphQL type"
            )

        if type_def.name.value != "Subscription":
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing "
                f"GraphQL definition for 'type {type_def.name.value}' "
                "(expected 'type Subscription')"
            )

        return cast(ObjectNodeType, type_def)

    @classmethod
    def __get_subscribers__(cls):
        aliases = cls.__aliases__ or {}
        defined_subscribers = cls.__get_defined_subscribers__()

        used_subscribers = []
        subscribers = {}

        for field_name in cls.graphql_fields:
            if aliases and field_name in aliases:
                subscription_name = aliases[field_name]
                if subscription_name in defined_subscribers:
                    used_subscribers.append(subscription_name)
                    subscribers[field_name] = defined_subscribers[subscription_name]

            elif field_name in defined_subscribers:
                used_subscribers.append(field_name)
                subscribers[field_name] = defined_subscribers[field_name]

        unused_subscribers = [
            f"resolve_{field_name}"
            for field_name in set(defined_subscribers) - set(used_subscribers)
        ]
        if unused_subscribers:
            raise ValueError(
                f"{cls.__name__} class was defined with subscribers for fields "
                f"not in  GraphQL type: {', '.join(unused_subscribers)}"
            )

        return subscribers

    @classmethod
    def __get_defined_subscribers__(cls) -> Dict[str, Callable]:
        resolvers = {}
        for name in dir(cls):
            if not name.startswith("subscribe_"):
                continue

            value = getattr(cls, name)
            if callable(value):
                resolvers[name[10:]] = value

        return resolvers

    @classmethod
    def __bind_to_schema__(cls, schema: GraphQLSchema):
        graphql_type = cast(GraphQLObjectType, schema.type_map[cls.graphql_name])

        for field_name, field_resolver in cls.resolvers.items():
            graphql_type.fields[field_name].resolve = field_resolver

        for field_name, field_subscriber in cls.subscribers.items():
            graphql_type.fields[field_name].subscribe = field_subscriber



from typing import Dict, Iterable, List, Tuple, Type, cast

from ariadne import (
    SchemaDirectiveVisitor,
    set_default_enum_values_on_schema,
    validate_schema_enum_values,
)
from graphql import (
    ConstDirectiveNode,
    DocumentNode,
    FieldDefinitionNode,
    GraphQLSchema,
    NamedTypeNode,
    ObjectTypeDefinitionNode,
    assert_valid_schema,
    build_ast_schema,
    concat_ast,
    parse,
)
from graphql.language import ast

from .base_type import BaseType
from .deferred_type import DeferredType
from .enum_type import EnumType

ROOT_TYPES = ["Query", "Mutation", "Subscription"]


def make_executable_schema(
    *types,
    merge_roots: bool = True,
):
    all_types: List[Type[BaseType]] = []
    find_requirements(all_types, types)

    real_types = [type_ for type_ in all_types if not isinstance(type_, DeferredType)]
    validate_no_missing_types(real_types, all_types)

    schema = build_schema(real_types, merge_roots)
    set_default_enum_values_on_schema(schema)
    assert_valid_schema(schema)
    validate_schema_enum_values(schema)
    repair_default_enum_values(schema, real_types)

    add_directives_to_schema(schema, real_types)

    return schema


def find_requirements(
    types_list: List[Type[BaseType]], types: Iterable[Type[BaseType]]
):
    for type_ in types:
        if type_ not in types_list:
            types_list.append(type_)

        find_requirements(types_list, type_.__requires__)


def validate_no_missing_types(
    real_types: List[Type[BaseType]], all_types: List[Type[BaseType]]
):
    deferred_names = [
        deferred.graphql_name
        for deferred in all_types
        if isinstance(deferred, DeferredType)
    ]

    real_names = [type_.graphql_name for type_ in real_types]
    missing_names = set(deferred_names) - set(real_names)
    if missing_names:
        raise ValueError(
            "Following types are defined as deferred and are missing "
            f"from schema: {', '.join(missing_names)}"
        )


def build_schema(
    types_list: List[Type[BaseType]], merge_roots: bool = True
) -> GraphQLSchema:
    schema_definitions: List[ast.DocumentNode] = []
    if merge_roots:
        schema_definitions.append(build_root_schema(types_list))
        for type_ in types_list:
            if type_.graphql_name not in ROOT_TYPES or not merge_roots:
                schema_definitions.append(parse(type_.__schema__))

    ast_document = concat_ast(schema_definitions)
    schema = build_ast_schema(ast_document)

    for type_ in types_list:
        type_.__bind_to_schema__(schema)

    return schema


def build_root_schema(types_list: List[Type[BaseType]]) -> DocumentNode:
    root_types: Dict[str, List[Type[BaseType]]] = {
        "Query": [],
        "Mutation": [],
        "Subscription": [],
    }

    for type_ in types_list:
        if type_.graphql_name in root_types:
            root_types[type_.graphql_name].append(type_)

    schema: List[DocumentNode] = []
    for types_defs in root_types.values():
        if len(types_defs) == 1:
            schema.append(parse(types_defs[0].__schema__))
        elif types_defs:
            schema.append(merge_root_types(types_defs))

    return concat_ast(schema)


def merge_root_types(types_list: List[Type[BaseType]]) -> DocumentNode:
    interfaces: List[NamedTypeNode] = []
    directives: List[ConstDirectiveNode] = []
    fields: Dict[str, Tuple[FieldDefinitionNode, Type[BaseType]]] = {}

    for type_ in types_list:
        type_definition = cast(
            ObjectTypeDefinitionNode,
            parse(type_.__schema__).definitions[0],
        )
        interfaces.extend(type_definition.interfaces)
        directives.extend(type_definition.directives)

        for field_def in type_definition.fields:
            field_name = field_def.name.value
            if field_name in fields:
                other_type_name = fields[field_name][1].__name__
                raise ValueError(
                    f"Multiple {type_.graphql_name} types are defining same field "
                    f"'{field_name}': {other_type_name}, {type_.__name__}"
                )

            fields[field_name] = (field_def, type_)

    merged_definition = ast.ObjectTypeDefinitionNode()
    merged_definition.name = ast.NameNode()
    merged_definition.name.value = types_list[0].graphql_name
    merged_definition.interfaces = tuple(interfaces)
    merged_definition.directives = tuple(directives)
    merged_definition.fields = tuple(
        fields[field_name][0] for field_name in sorted(fields)
    )

    merged_document = DocumentNode()
    merged_document.definitions = (merged_definition,)

    return merged_document


def add_directives_to_schema(schema: GraphQLSchema, types_list: List[Type[BaseType]]):
    directives: Dict[str, Type[SchemaDirectiveVisitor]] = {}
    for type_ in types_list:
        visitor = getattr(type_, "__visitor__", None)
        if visitor and issubclass(visitor, SchemaDirectiveVisitor):
            directives[type_.graphql_name] = visitor

    if directives:
        SchemaDirectiveVisitor.visit_schema_directives(schema, directives)


def repair_default_enum_values(schema, types_list: List[Type[BaseType]]) -> None:
    for type_ in types_list:
        if issubclass(type_, EnumType):
            type_.__bind_to_default_values__(schema)



from typing import Dict, Type, Union, cast

from graphql import (
    DefinitionNode,
    GraphQLFieldResolver,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
)

from .base_type import BaseType
from .dependencies import Dependencies, get_dependencies_from_object_type
from .resolvers_mixin import ResolversMixin
from .types import FieldsDict, RequirementsDict
from .utils import parse_definition

ObjectNodeType = Union[ObjectTypeDefinitionNode, ObjectTypeExtensionNode]


class ObjectType(BaseType, ResolversMixin):
    __abstract__ = True

    graphql_name: str
    graphql_type: Union[Type[ObjectTypeDefinitionNode], Type[ObjectTypeExtensionNode]]

    resolvers: Dict[str, GraphQLFieldResolver]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)
        cls.graphql_fields = cls.__get_fields__(graphql_def)

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

        dependencies = cls.__get_dependencies__(graphql_def)
        cls.__validate_requirements__(requirements, dependencies)

        if callable(cls.__fields_args__):
            cls.__fields_args__ = cls.__fields_args__(cls.graphql_fields, True)

        cls.__validate_fields_args__()

        if callable(cls.__aliases__):
            cls.__aliases__ = cls.__aliases__(cls.graphql_fields)

        cls.__validate_aliases__()
        cls.resolvers = cls.__get_resolvers__()

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> ObjectNodeType:
        if not isinstance(
            type_def, (ObjectTypeDefinitionNode, ObjectTypeExtensionNode)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ without GraphQL type"
            )

        if type_def.name.value == "Subscription":
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing "
                f"GraphQL definition for 'type Subscription' which is only supported "
                "by subsclassess of 'SubscriptionType'"
            )

        return cast(ObjectNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: ObjectNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, ObjectTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} graphql type was defined without required GraphQL "
                f"type definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != ObjectTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL type "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __get_fields__(cls, type_def: ObjectNodeType) -> FieldsDict:
        if not type_def.fields and not (
            isinstance(type_def, ObjectTypeExtensionNode)
            and (type_def.directives or type_def.interfaces)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing "
                f"empty GraphQL type definition"
            )

        return {field.name.value: field for field in type_def.fields}

    @classmethod
    def __get_dependencies__(cls, type_def: ObjectNodeType) -> Dependencies:
        return get_dependencies_from_object_type(type_def)

    @classmethod
    def __bind_to_schema__(cls, schema):
        graphql_type = schema.type_map.get(cls.graphql_name)

        for field_name, field_resolver in cls.resolvers.items():
            graphql_type.fields[field_name].resolve = field_resolver



from graphql import ObjectTypeDefinitionNode

from .base_type import BaseType


class DeferredType(BaseType):
    graphql_type = ObjectTypeDefinitionNode

    def __init__(self, name: str):
        self.graphql_name = name

    @classmethod
    def __bind_to_schema__(cls, *_):
        raise NotImplementedError("DeferredType cannot be bound to schema")



from typing import Type, Union, cast

from graphql import (
    DefinitionNode,
    GraphQLTypeResolver,
    GraphQLSchema,
    GraphQLUnionType,
    UnionTypeDefinitionNode,
    UnionTypeExtensionNode,
)

from .base_type import BaseType
from .dependencies import Dependencies, get_dependencies_from_union_type
from .types import RequirementsDict
from .utils import parse_definition

UnionNodeType = Union[UnionTypeDefinitionNode, UnionTypeExtensionNode]


class UnionType(BaseType):
    __abstract__ = True

    graphql_type: Union[Type[UnionTypeDefinitionNode], Type[UnionTypeExtensionNode]]
    resolve_type: GraphQLTypeResolver

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

        dependencies = cls.__get_dependencies__(graphql_def)
        cls.__validate_requirements__(requirements, dependencies)

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> UnionNodeType:
        if not isinstance(type_def, (UnionTypeDefinitionNode, UnionTypeExtensionNode)):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ "
                "without GraphQL union"
            )

        return cast(UnionNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: UnionNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, UnionTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} class was defined without required GraphQL union "
                f"definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != UnionTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL union "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __get_dependencies__(cls, type_def: UnionNodeType) -> Dependencies:
        return get_dependencies_from_union_type(type_def)

    @classmethod
    def __bind_to_schema__(cls, schema: GraphQLSchema):
        graphql_type = cast(GraphQLUnionType, schema.type_map.get(cls.graphql_name))
        graphql_type.resolve_type = cls.resolve_type



from typing import Any, Mapping

from graphql import (
    DefinitionNode,
    GraphQLResolveInfo,
    ListTypeNode,
    NonNullTypeNode,
    TypeNode,
    parse,
)


def parse_definition(type_name: str, schema: Any) -> DefinitionNode:
    if not isinstance(schema, str):
        raise TypeError(
            f"{type_name} class was defined with __schema__ of invalid type: "
            f"{type(schema).__name__}"
        )

    definitions = parse(schema).definitions

    if len(definitions) > 1:
        definitions_types = [type(definition).__name__ for definition in definitions]
        raise ValueError(
            f"{type_name} class was defined with __schema__ containing more "
            f"than one GraphQL definition (found: {', '.join(definitions_types)})"
        )

    return definitions[0]


def unwrap_type_node(field_type: TypeNode):
    if isinstance(field_type, (NonNullTypeNode, ListTypeNode)):
        return unwrap_type_node(field_type.type)
    return field_type


def create_alias_resolver(field_name: str):
    def default_aliased_field_resolver(
        source: Any, info: GraphQLResolveInfo, **args: Any
    ) -> Any:
        value = (
            source.get(field_name)
            if isinstance(source, Mapping)
            else getattr(source, field_name, None)
        )

        if callable(value):
            return value(info, **args)
        return value

    return default_aliased_field_resolver



from typing import Callable, Dict, Optional, Union, cast

from graphql import (
    DefinitionNode,
    InputObjectTypeDefinitionNode,
    InputObjectTypeExtensionNode,
)

from .base_type import BaseType
from .dependencies import Dependencies, get_dependencies_from_input_type
from .types import InputFieldsDict, RequirementsDict
from .utils import parse_definition

Args = Dict[str, str]
InputNodeType = Union[InputObjectTypeDefinitionNode, InputObjectTypeExtensionNode]


class InputType(BaseType):
    __abstract__ = True
    __args__: Optional[Union[Args, Callable[..., Args]]] = None

    graphql_fields: InputFieldsDict

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)
        cls.graphql_fields = cls.__get_fields__(graphql_def)

        if callable(cls.__args__):
            # pylint: disable=not-callable
            cls.__args__ = cls.__args__(cls.graphql_fields)

        cls.__validate_args__()

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

        dependencies = cls.__get_dependencies__(graphql_def)
        cls.__validate_requirements__(requirements, dependencies)

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> InputNodeType:
        if not isinstance(
            type_def, (InputObjectTypeDefinitionNode, InputObjectTypeExtensionNode)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ without GraphQL input"
            )

        return cast(InputNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: InputNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, InputObjectTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} graphql type was defined without required GraphQL "
                f"type definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != InputObjectTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL input "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __get_fields__(cls, type_def: InputNodeType) -> InputFieldsDict:
        if not type_def.fields and not (
            isinstance(type_def, InputObjectTypeExtensionNode) and type_def.directives
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing empty "
                f"GraphQL input definition"
            )

        return {field.name.value: field for field in type_def.fields}

    @classmethod
    def __validate_args__(cls):
        if not cls.__args__:
            return

        invalid_args = set(cls.__args__) - set(cls.graphql_fields)
        if invalid_args:
            raise ValueError(
                f"{cls.__name__} class was defined with args for fields not in "
                f"GraphQL input: {', '.join(invalid_args)}"
            )

    @classmethod
    def __get_dependencies__(cls, type_def: InputNodeType) -> Dependencies:
        return get_dependencies_from_input_type(type_def)

    @classmethod
    def __bind_to_schema__(cls, schema):
        if not cls.__args__:
            return

        graphql_type = schema.type_map.get(cls.graphql_name)
        for field_name, field_target in cls.__args__.items():
            graphql_type.fields[field_name].out_name = field_target



from typing import Tuple, Union, Set

from graphql import (
    ConstDirectiveNode,
    FieldDefinitionNode,
    InputObjectTypeDefinitionNode,
    InputObjectTypeExtensionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    InterfaceTypeExtensionNode,
    NamedTypeNode,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
    UnionTypeDefinitionNode,
    UnionTypeExtensionNode,
)

from .utils import unwrap_type_node

GRAPHQL_TYPES = ("ID", "Int", "String", "Boolean", "Float")

Dependencies = Tuple[str, ...]


def get_dependencies_from_object_type(
    graphql_type: Union[
        InterfaceTypeDefinitionNode,
        InterfaceTypeExtensionNode,
        ObjectTypeDefinitionNode,
        ObjectTypeExtensionNode,
    ]
) -> Dependencies:
    dependencies: Set[str] = set()
    dependencies.update(
        get_dependencies_from_directives(graphql_type.directives),
        get_dependencies_from_fields(graphql_type.fields),
        get_dependencies_from_interfaces(graphql_type.interfaces),
    )

    if graphql_type.name.value in dependencies:
        # Remove self-dependency
        dependencies.remove(graphql_type.name.value)

    return tuple(dependencies)


def get_dependencies_from_input_type(
    graphql_type: Union[InputObjectTypeDefinitionNode, InputObjectTypeExtensionNode]
) -> Dependencies:
    dependencies: Set[str] = set()
    dependencies.update(
        get_dependencies_from_directives(graphql_type.directives),
        get_dependencies_from_input_fields(graphql_type.fields),
    )

    if graphql_type.name.value in dependencies:
        # Remove self-dependency
        dependencies.remove(graphql_type.name.value)

    return tuple(dependencies)


def get_dependencies_from_union_type(
    graphql_type: Union[UnionTypeDefinitionNode, UnionTypeExtensionNode]
) -> Dependencies:
    dependencies: Set[str] = set()
    dependencies.update(
        get_dependencies_from_directives(graphql_type.directives),
        get_dependencies_from_interfaces(graphql_type.types),
    )

    if graphql_type.name.value in dependencies:
        # Remove self-dependency
        dependencies.remove(graphql_type.name.value)

    return tuple(dependencies)


def get_dependencies_from_directives(
    directives: Tuple[ConstDirectiveNode, ...]
) -> Dependencies:
    dependencies: Set[str] = set()
    for directive in directives:
        dependencies.add(directive.name.value)
    return tuple(dependencies)


def get_dependencies_from_fields(
    fields: Tuple[FieldDefinitionNode, ...]
) -> Dependencies:
    dependencies: Set[str] = set()

    for field_def in fields:
        dependencies.update(get_dependencies_from_directives(field_def.directives))

        # Get dependency from return type
        field_type = unwrap_type_node(field_def.type)
        if isinstance(field_type, NamedTypeNode):
            field_type_name = field_type.name.value
            if field_type_name not in GRAPHQL_TYPES:
                dependencies.add(field_type_name)

        # Get dependency from arguments
        for arg_def in field_def.arguments:
            dependencies.update(get_dependencies_from_directives(arg_def.directives))

            arg_type = unwrap_type_node(arg_def.type)
            if isinstance(arg_type, NamedTypeNode):
                arg_type_name = arg_type.name.value
                if arg_type_name not in GRAPHQL_TYPES:
                    dependencies.add(arg_type_name)

    return tuple(dependencies)


def get_dependencies_from_input_fields(
    fields: Tuple[InputValueDefinitionNode, ...]
) -> Dependencies:
    dependencies: Set[str] = set()

    for field_def in fields:
        dependencies.update(get_dependencies_from_directives(field_def.directives))

        # Get dependency from return type
        field_type = unwrap_type_node(field_def.type)
        if isinstance(field_type, NamedTypeNode):
            field_type_name = field_type.name.value
            if field_type_name not in GRAPHQL_TYPES:
                dependencies.add(field_type_name)

    return tuple(dependencies)


def get_dependencies_from_interfaces(
    interfaces: Tuple[NamedTypeNode, ...]
) -> Dependencies:
    dependencies: Set[str] = set()
    for interface in interfaces:
        dependencies.add(interface.name.value)
    return tuple(dependencies)



from typing import Dict, Callable, Type, Union, cast

from graphql import (
    DefinitionNode,
    GraphQLFieldResolver,
    GraphQLInterfaceType,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLTypeResolver,
    InterfaceTypeDefinitionNode,
    InterfaceTypeExtensionNode,
)

from ariadne import type_implements_interface

from .base_type import BaseType
from .dependencies import Dependencies, get_dependencies_from_object_type
from .resolvers_mixin import ResolversMixin
from .types import FieldsDict, RequirementsDict
from .utils import parse_definition

InterfaceNodeType = Union[InterfaceTypeDefinitionNode, InterfaceTypeExtensionNode]


class InterfaceType(BaseType, ResolversMixin):
    __abstract__ = True

    graphql_name: str
    graphql_type: Union[
        Type[InterfaceTypeDefinitionNode], Type[InterfaceTypeExtensionNode]
    ]

    resolve_type: GraphQLTypeResolver
    resolvers: Dict[str, GraphQLFieldResolver]

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)
        cls.graphql_fields = cls.__get_fields__(graphql_def)

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

        dependencies = cls.__get_dependencies__(graphql_def)
        cls.__validate_requirements__(requirements, dependencies)

        if callable(cls.__fields_args__):
            cls.__fields_args__ = cls.__fields_args__(cls.graphql_fields, True)

        cls.__validate_fields_args__()

        if callable(cls.__aliases__):
            cls.__aliases__ = cls.__aliases__(cls.graphql_fields)

        cls.__validate_aliases__()
        cls.resolvers = cls.__get_resolvers__()

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> InterfaceNodeType:
        if not isinstance(
            type_def, (InterfaceTypeDefinitionNode, InterfaceTypeExtensionNode)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ without "
                "GraphQL interface"
            )

        return cast(InterfaceNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: InterfaceNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, InterfaceTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} graphql type was defined without required GraphQL "
                f"type definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != InterfaceTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL interface "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __get_fields__(cls, type_def: InterfaceNodeType) -> FieldsDict:
        if not type_def.fields and not (
            isinstance(type_def, InterfaceTypeExtensionNode)
            and (type_def.directives or type_def.interfaces)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing empty "
                f"GraphQL interface definition"
            )

        return {field.name.value: field for field in type_def.fields}

    @classmethod
    def __get_defined_resolvers__(cls) -> Dict[str, Callable]:
        resolvers = super().__get_defined_resolvers__()
        resolvers.pop("type", None)
        return resolvers

    @classmethod
    def __get_dependencies__(cls, type_def: InterfaceNodeType) -> Dependencies:
        return get_dependencies_from_object_type(type_def)

    @classmethod
    def __bind_to_schema__(cls, schema: GraphQLSchema):
        graphql_type = cast(GraphQLInterfaceType, schema.type_map.get(cls.graphql_name))
        graphql_type.resolve_type = cls.resolve_type

        for type_ in schema.type_map.values():
            if not type_implements_interface(cls.graphql_name, type_):
                continue

            type_ = cast(GraphQLObjectType, type_)
            for field_name, field_resolver in cls.resolvers.items():
                if not type_.fields[field_name].resolve:
                    type_.fields[field_name].resolve = field_resolver



from typing import Dict, Type

from graphql import (
    DefinitionNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
)

FieldsDict = Dict[str, FieldDefinitionNode]
InputFieldsDict = Dict[str, InputValueDefinitionNode]
RequirementsDict = Dict[str, Type[DefinitionNode]]



from typing import Dict, Optional, Type, Union, cast

from graphql import (
    DefinitionNode,
    FieldDefinitionNode,
    GraphQLFieldResolver,
    ObjectTypeDefinitionNode,
    ObjectTypeExtensionNode,
)

from .base_type import BaseType
from .dependencies import Dependencies, get_dependencies_from_object_type
from .types import RequirementsDict
from .utils import parse_definition

ObjectNodeType = Union[ObjectTypeDefinitionNode, ObjectTypeExtensionNode]


class MutationType(BaseType):
    __abstract__ = True
    __args__: Optional[Dict[str, str]] = None

    graphql_name = "Mutation"
    graphql_type: Union[Type[ObjectTypeDefinitionNode], Type[ObjectTypeExtensionNode]]

    mutation_name: str
    resolve_mutation: GraphQLFieldResolver

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)

        field = cls.__get_field__(graphql_def)
        cls.mutation_name = field.name.value

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

        dependencies = cls.__get_dependencies__(graphql_def)
        cls.__validate_requirements__(requirements, dependencies)

        cls.__validate_args__(field)
        cls.__validate_resolve_mutation__()

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> ObjectNodeType:
        if not isinstance(
            type_def, (ObjectTypeDefinitionNode, ObjectTypeExtensionNode)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ without GraphQL type"
            )

        if type_def.name.value != "Mutation":
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing "
                f"GraphQL definition for 'type {type_def.name.value}' while "
                "'type Mutation' was expected"
            )

        return cast(ObjectNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: ObjectNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, ObjectTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} graphql type was defined without required GraphQL "
                f"type definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != ObjectTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL type "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __get_field__(cls, type_def: ObjectNodeType) -> FieldDefinitionNode:
        if not type_def.fields:
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ containing "
                f"empty GraphQL type definition"
            )

        if len(type_def.fields) != 1:
            raise ValueError(
                f"{cls.__name__} class subclasses 'MutationType' class which "
                "requires __schema__ to define exactly one field"
            )

        return type_def.fields[0]

    @classmethod
    def __get_dependencies__(cls, type_def: ObjectNodeType) -> Dependencies:
        return get_dependencies_from_object_type(type_def)

    @classmethod
    def __validate_args__(cls, field: FieldDefinitionNode):
        if not cls.__args__:
            return

        field_args = [arg.name.value for arg in field.arguments]
        invalid_args = set(cls.__args__) - set(field_args)
        if invalid_args:
            raise ValueError(
                f"{cls.__name__} class was defined with args not on "
                f"'{field.name.value}' GraphQL field: {', '.join(invalid_args)}"
            )

    @classmethod
    def __validate_resolve_mutation__(cls):
        resolver = getattr(cls, "resolve_mutation", None)
        if not resolver:
            raise AttributeError(
                f"{cls.__name__} class was defined without required "
                "'resolve_mutation' attribute"
            )

        if not callable(resolver):
            raise TypeError(
                f"{cls.__name__} class was defined with attribute "
                "'resolve_mutation' but it's not callable"
            )

    @classmethod
    def __bind_to_schema__(cls, schema):
        graphql_type = schema.type_map.get(cls.graphql_name)
        graphql_type.fields[cls.mutation_name].resolve = cls.resolve_mutation

        if cls.__args__:
            field_args = graphql_type.fields[cls.mutation_name].args
            for arg_name, out_name in cls.__args__.items():
                field_args[arg_name].out_name = out_name



from typing import Optional, Type, Union, cast

from graphql import (
    DefinitionNode,
    GraphQLScalarSerializer,
    GraphQLScalarType,
    GraphQLScalarLiteralParser,
    GraphQLScalarValueParser,
    GraphQLSchema,
    ScalarTypeDefinitionNode,
    ScalarTypeExtensionNode,
)

from .base_type import BaseType
from .types import RequirementsDict
from .utils import parse_definition

ScalarNodeType = Union[ScalarTypeDefinitionNode, ScalarTypeExtensionNode]


class ScalarType(BaseType):
    __abstract__ = True

    graphql_type: Union[Type[ScalarTypeDefinitionNode], Type[ScalarTypeExtensionNode]]

    serialize: Optional[GraphQLScalarSerializer] = None
    parse_value: Optional[GraphQLScalarValueParser] = None
    parse_literal: Optional[GraphQLScalarLiteralParser] = None

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if cls.__dict__.get("__abstract__"):
            return

        cls.__abstract__ = False

        graphql_def = cls.__validate_schema__(
            parse_definition(cls.__name__, cls.__schema__)
        )

        cls.graphql_name = graphql_def.name.value
        cls.graphql_type = type(graphql_def)

        requirements = cls.__get_requirements__()
        cls.__validate_requirements_contain_extended_type__(graphql_def, requirements)

    @classmethod
    def __validate_schema__(cls, type_def: DefinitionNode) -> ScalarNodeType:
        if not isinstance(
            type_def, (ScalarTypeDefinitionNode, ScalarTypeExtensionNode)
        ):
            raise ValueError(
                f"{cls.__name__} class was defined with __schema__ "
                "without GraphQL scalar"
            )

        return cast(ScalarNodeType, type_def)

    @classmethod
    def __validate_requirements_contain_extended_type__(
        cls, type_def: ScalarNodeType, requirements: RequirementsDict
    ):
        if not isinstance(type_def, ScalarTypeExtensionNode):
            return

        graphql_name = type_def.name.value
        if graphql_name not in requirements:
            raise ValueError(
                f"{cls.__name__} graphql type was defined without required GraphQL "
                f"scalar, definition for '{graphql_name}' in __requires__"
            )

        if requirements[graphql_name] != ScalarTypeDefinitionNode:
            raise ValueError(
                f"{cls.__name__} requires '{graphql_name}' to be GraphQL scalar "
                f"but other type was provided in '__requires__'"
            )

    @classmethod
    def __bind_to_schema__(cls, schema: GraphQLSchema):
        graphql_type = cast(GraphQLScalarType, schema.type_map.get(cls.graphql_name))

        # See mypy bug https://github.com/python/mypy/issues/2427
        if cls.serialize:
            graphql_type.serialize = cls.serialize  # type: ignore
        if cls.parse_value:
            graphql_type.parse_value = cls.parse_value  # type: ignore
        if cls.parse_literal:
            graphql_type.parse_literal = cls.parse_literal  # type: ignore
